# Isothermal vapor hydraulic conductivity
import numpy as np


def init_model_k_vapor(self):
    # Physical parameters
    self.temperature = 20  # temperature in degree Celcius
    self.water_density = 1000  # water density in kg / m^3
    self.gravity = 9.81  # gravitational acceleration in m / s^2
    # Conversion factors of input data
    self.convert_h = 0.01  # convert pressure head to m H2O
    self.convert_theta = 1  # convert water content to volumetric fraction
    self.convert_k = 0.01  # convert hydraulic conductivity to m/s


def model_k_vapor(self, h, theta, theta_s):
    """Isothermal vapor hydraulic conductivity

    as a function of h, θ, θs, defined in Saito et al. (2006)
    https://doi.org/10.2136/vzj2006.0007

    === How to use (example code) ===

    import unsatfit
    f = unsatfit.Fit()
    # Set temperature and conversion factors of input data
    f.temperature = 20  # temperature in degree Celcius
    f.convert_h = 0.01  # convert pressure head to m H2O
    f.convert_theta = 1  # convert water content to volumetric fraction
    f.convert_k = 0.01  # convert hydraulic conductivity to m/s
    # Set "k_vapor=True" when using set_model()
    # Then vapor component is added to hydraulic conductivity function
    f.set_model('DVC', const=[f.get_wrf_vg2ch()], k_vapor=True)
    """
    # Physical constant in SI unit
    water_mole = 0.018015  # kg / mol
    gas_constant = 8.314462  # J / mol K
    # Conversion
    centigrade_offset = 273.15
    t = self.temperature + centigrade_offset
    h = h * self.convert_h
    theta_s = theta_s * self.convert_theta
    theta = theta * self.convert_theta
    # Calculation
    sat_vap_density = 0.001 * \
        np.exp(31.3716 - 6014.79 / t - 0.00792495 * t) / t
    diff_water_air = 0.0000214 * (t / centigrade_offset) ** 2
    air_porosity = theta_s - theta
    tortuosity = air_porosity ** (7 / 3) / (theta_s ** 2)
    vapor_diffusivity = tortuosity * air_porosity * diff_water_air
    mgrt = water_mole * self.gravity / (gas_constant * t)  # Mg/RT
    relative_humidity = np.exp(-h * mgrt)
    k = vapor_diffusivity * sat_vap_density / \
        self.water_density * mgrt * relative_humidity
    k = k / self.convert_k
    return k



def contour(self, x_name, y_name):
    """Contour of RMSE"""
    import numpy as np
    import matplotlib.pyplot as plt
    from .unsatfit import Fit

    # Set data and residual function
    param = self.model[self.model_name.replace('Modified ', '')]['param']
    if self.ht_only:
        data = self.swrc
        par = list(self.fitted)
        for c in self.const_ht:
            par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
        for i in sorted(self.model_k_only, reverse=True):
            param = param[:i] + param[i + 1:]
    else:
        data = self.unsat
        par = list(self.fitted)
        for c in self.const:
            par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    const = []
    for i in range(len(par)):
        const.append([i + 1, par[i]])
    xi = param.index(x_name)
    yi = param.index(y_name)
    x = par[xi]
    y = par[yi]
    for i in sorted((xi, yi), reverse=True):
        const = const[:i] + const[i + 1:]

    f = Fit()
    if 'Modified' in self.model_name:
        f.set_model(self.model_name.replace('Modified ', ''), const=const)
        f.modified_model(self.hs)
    else:
        f.set_model(self.model_name, const=const)
    if self.ht_only:
        residual = f.residual_ht
    else:
        residual = f.residual_log10_hk

    # Make grid of (X,Y)
    x_min, x_max = self.contour_range_x
    y_min, y_max = self.contour_range_y
    x_delta = (x_max - x_min) / self.contour_smooth
    y_delta = (y_max - y_min) / self.contour_smooth
    x_grid = np.arange(x * x_min, x * x_max, x * x_delta)
    y_grid = np.arange(y * y_min, y * y_max, y * y_delta)
    X, Y = np.meshgrid(x_grid, y_grid)

    # Calculate RMSE for (X,Y) as Z
    z = []
    for yi in y_grid:
        row = []
        for xi in x_grid:
            mse = np.average(residual((xi, yi), *data)**2)
            row.append(mse ** 0.5)
        z.append(row)
    Z = np.array(z)

    # Draw contour of (X,Y,Z)
    fig, ax = plt.subplots(figsize=[self.fig_width, self.fig_height])
    fig.subplots_adjust(top=1 - self.top_margin, bottom=self.bottom_margin,
                        right=1 - self.right_margin, left=self.left_margin)
    CS = ax.contour(X, Y, Z, self.contour_level, colors=self.contour_color)
    ax.set_xlabel(self.label(x_name))
    ax.set_ylabel(self.label(y_name))
    ax.clabel(CS, inline=True, fontsize=10)

    if self.contour_show_marker:
        ax.plot(x, y, color=self.contour_marker_color,
                marker='o', linestyle='')

    if self.save_fig:
        plt.savefig(self.filename)
    if self.show_fig:
        plt.show()


def label(self, name):
    label = '$' + name + '$'
    for i in [['qs', '\\theta_s'], ['qr', '\\theta_r'], ['hb', 'h_b'], ['hm2', 'h_{m2}'], [
            'hm', 'h_m'], ['Ks', 'K_s'], ['sigma', '\\sigma'], ['1', '_1'], ['2', '_2']]:
        label = label.replace(i[0], i[1])
    return label



# Initialize and set models

def init_model(self):
    self.model = {}
    self.init_model_bc()
    self.init_model_vg()
    self.init_model_ln()
    self.init_model_fx()
    self.init_model_bc2()
    self.init_model_vg2()
    self.init_model_ln2()
    self.init_model_vgbc()
    self.init_model_kobc()
    self.init_model_fs()
    self.init_model_pe()
    self.init_model_tri_vg()
    self.init_model_bvv()
    self.init_model_vvp()
    self.init_model_k_vapor()
    self.output_format = {
        'qs': '.3f', 'qr': '.3f', 'qa': '.3f', 'w1': '.3f', 'ww2': '.3f', 'a': '.3', 'a1': '.3', 'a2': '.3', 'a3': '.3',
        'm': '.3f', 'n': '.3f', 'm1': '.3f', 'm2': '.3f', 'm3': '.3f', 'hm': '.2f', 'hm1': '.2f', 'hm2': '.2f',
        'sigma': '.3f', 'sigma1': '.3f', 'sigma2': '.3f', 'hb': '.2f', 'hb1': '.2f', 'hb2': '.2f', 'hf': '.3f',
        'hc': '.2f', 'l': '.3', 'l1': '.3', 'l2': '.3', 'he': '.2f',
        'Ks': '.2e', 'p': '.3f', 'p1': '.3f', 'p2': '.3f', 'q': '.3f', 'q1': '.3f', 'q2': '.3f', 'r': '.3f', 'omega': '.3'
    }
    self.r2_format = '.3f'


def set_model(self, model, const=[], k_vapor=False):
    """Set model

    Parameters:

        model: name of the model
        const: constant parameters of the model. For example
            [[1,0.5], [2,0], [3,0.2]] sets 1st parameter as 0.5, 2nd parameter as 0 and 3rd parameter as 0.2
            [[1,0.5], 'qr=0']] sets 1st parameter as 0.5 and qr as 0
            [wrf, 'r=1'] where wrf is a tuple of all WRF parameters, sets the WRF parameters and r as 1
        k_vapor: if using isothermal vapor hydraulic conductivity
            When True, self.k_vapor() is added to hydraulic conductivity.
            Several parameters need to be set when using this feature.
            See comment in k_vapor() function.
    """
    self.model_name = model
    model = self.model[model]
    self.f_ht, self.f_hk = model['function']
    if k_vapor:
        def k_vapor(p, h):
            theta = self.f_ht(p, h)
            theta_s = self.f_ht(p, 0)
            k = model['function'][1](p, h)
            return k + self.model_k_vapor(h, theta, theta_s)
        self.f_hk = k_vapor
    self.b_func = model['bound']
    self.param = model['param']
    self.model_k_only = model['k-only']
    if 'sort_param' in model:
        self.sort_param = model['sort_param']
    else:
        self.sort_param = None
    # Set get_init() and get_wrf() functions
    self.set_get_init_wrf()
    # Organize parameters
    self.organize_parameters(const)


def set_get_init_wrf(self):
    """Set get_init() and get_wrf() functions"""
    model = self.model[self.model_name]

    def get_init_not_defined():
        print('get_init function not defined for {0} model'.format(
            self.model_name))
        exit(1)
    if 'get_init' in model:
        self.get_init = model['get_init']
    else:
        self.get_init = get_init_not_defined

    def get_wrf_not_defined():
        print('get_wrf function not defined for {0} model'.format(
            self.model_name))
        exit(1)
    if 'get_wrf' in model:
        self.get_wrf = model['get_wrf']
    else:
        self.get_wrf = get_wrf_not_defined


def organize_parameters(self, const):
    """Organize parameters"""
    # Recostruct const to allow alternative expressions
    model = self.model[self.model_name]
    self.reconstruct_const(const)
    # Calculate self.p_k_only
    self.k_only()
    # Calculate self.const_ht by eliminating K-only parameters
    self.const_ht = []
    for c in self.const:
        if c[0] - 1 not in self.model_k_only:
            self.const_ht.append(
                [c[0] - sum(1 for x in self.model_k_only if x < c[0] - 1), c[1]])
    # Calculate self.param and self.param_ht
    self.param_ht = self.param
    for i in sorted(self.model_k_only, reverse=True):
        self.param_ht = self.param_ht[:i] + self.param_ht[i + 1:]
    for c in sorted(self.const_ht, reverse=True):
        self.param_ht = self.param_ht[:c[0] - 1] + self.param_ht[c[0]:]
    for c in sorted(self.const, reverse=True):
        self.param = self.param[:c[0] - 1] + self.param[c[0]:]
    # Model description
    self.param_const = []
    self.value_const = []
    for i in self.const:
        self.param_const.append(model['param'][i[0] - 1])
        self.value_const.append(float(i[1]))
    self.const_description = self.format(
        self.param_const, ShowR2=False).format(*self.value_const)
    self.model_description = self.model_name + \
        ' model with ' + self.const_description


def reconstruct_const(self, const):
    """Reconstruct the list of parameter constraints for the model.

    The method interprets the input `const` (a list of constraints) in
    three possible formats and converts them into a standardized list
    of `[parameter_index, value]` pairs:

    1. String assignments (e.g., "q=1"):
       - Splits into parameter name and value.
       - Checks that the parameter exists in `self.param`.
       - Converts to `[index, float(value)]`.

    2. Full parameter list/tuple (longer than 2 elements):
       - Assumes the list provides values for all water retention
         parameters except those in `self.model_k_only`.
       - Matches given values to the correct parameter indices.
       - Errors out if the number of values does not match expectations.

    3. Explicit index/value pairs (e.g., `[2, 0]`):
       - Directly added to the reconstructed constraints.

    Finally, the reconstructed constraints are sorted and stored in
    `self.const`.

    Parameters
    ----------
    const : list
        A list of constraints expressed either as strings, full parameter
        lists, or explicit index/value pairs.

    Raises
    ------
    SystemExit
        If a parameter is not found in `self.param` or the number of
        provided values does not match the required number of parameters.
    """
    reconst = []
    for i in const:
        if '=' in str(i):  # expression like 'q=1'
            p, value = i.split('=')
            if p not in self.param:
                print('Parameter {0} not in this model'.format(p))
                exit(1)
            reconst.append([self.param.index(p) + 1, float(value)])
        elif len(i) > 2:  # tuple or list of all water retention parameters
            p = list(range(len(self.param)))
            for j in self.model_k_only:
                p.remove(j)
            if len(p) == len(i):
                k = 0
                for j in p:
                    reconst.append([j + 1, i[k]])
                    k += 1
            else:
                print(
                    '{0} parameters required for water retention function, but {1} parameters {2} given.'.format(len(p), len(i), i))
                exit()
        else:  # expression like [2, 0]
            reconst.append(i)
    self.const = sorted(reconst)


def k_only(self):
    """Compute the reindexed positions of k-only parameters after removing constants.

    Reads:
        - self.model_k_only (list[int]):
            0-based indices of parameters that belong to the k-only group in the
            full parameter vector.
        - self.const (list[list]):
            Each element is [i, _] where i is a 1-based index of a parameter
            that is fixed (constant).

    Procedure:
        1) Drop any indices from self.model_k_only that are listed as constants.
        2) For each remaining index k, subtract the number of constant indices
            that are strictly less than k (counting duplicates if present).
            This yields the index k would have after removing all constant
            positions from the full parameter vector.
        3) Store the resulting indices in descending order in self.p_k_only.

    Produces:
        - self.p_k_only (list[int]):
            Reindexed positions of k-only parameters in the reduced (constants-removed)
            parameter vector, sorted in descending order.
    """
    # Convert 1-based constant indices to 0-based and sort ascending
    const0 = sorted([c[0] - 1 for c in self.const])
    # Remove entries that are constants from model_k_only
    remain = [k for k in self.model_k_only if k not in const0]
    # Re-map each remaining index to its position after removing all constants
    res = []
    for k in remain:
        # Count how many constant positions occur strictly before k
        shift = 0
        for c in const0:
            if c < k:
                shift += 1
        # New index after removing the preceding constants
        res.append(k - shift)
    # Store in descending order as required
    self.p_k_only = sorted(res, reverse=True)


def test_k_only(self):
    self.test_k_only_unit('VG', (1, 2), [5, 3, 2])
    self.test_k_only_unit('VG', (1, 2, 3), [4, 2, 1])
    self.test_k_only_unit('VG', (1, 2, 3, 7), [3, 2, 1])
    self.test_k_only_unit('VG', (1, 2, 3, 4), [3, 1, 0])
    self.test_k_only_unit('VG', (7,), [6, 5, 4])
    self.test_k_only_unit('VG', (1, 2, 8), [3, 2])
    self.test_k_only_unit('VG', (1, 2, 3, 4, 6), [2, 0])
    self.test_k_only_unit('VG', (1, 3, 4, 6), [3, 1])
    self.test_k_only_unit('VG', (1, 3, 4, 5), [3, 1])


def test_k_only_unit(self, model, const, result):
    import json
    import sys
    from .unsatfit import Fit
    const_par = []
    for c in const:
        const_par.append([c, 0])
    f = Fit()
    f.swrc = (1, 2, 3)
    f.unsat = (2, 3, 4)
    f.set_model(model, const=const_par)
    f.k_only()
    if json.dumps(f.p_k_only) == json.dumps(result):
        if self.debug:
            print(
                f'success: model_k_only = {f.model_k_only} const = {const} p_k_only = {f.p_k_only}')
    else:
        print(
            f'Test failed with test_k_only_unit().\nk_only = {f.model_k_only} const = {const} p_k_only = {f.p_k_only} expected {result}')
        sys.exit()



# Modified model

import numpy as np


def modified_model(self, hs):
    """Modified model introducing air entry head
    Call modified_model(hs) after set_model()

    Requirement: param starts with (qs, qr) and k_only starts with Ks
    """
    import copy
    self.hs = hs
    self.f_ht_org = copy.deepcopy(self.f_ht)
    self.f_hk_org = copy.deepcopy(self.f_hk)
    self.f_ht = self.modified_ht
    self.f_hk = self.modified_hk
    self.model_name = "Modified " + self.model_name
    self.model_description = self.model_name + \
        ' model (hs = ' + str(hs) + ') with ' + self.const_description


def modified_ht(self, p, x):
    par = list(p)
    for c in self.const_ht:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr = par[:2]
    modify = (qs - qr) / (self.f_ht_org(p, self.hs) - qr)
    se = (self.f_ht_org(p, x) - qr) / (qs - qr)
    return np.where(x < self.hs, qs, modify * se * (qs - qr) + qr)


def modified_hk(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    ks = par[self.model_k_only[0]]
    modify = ks / self.f_hk_org(p, self.hs)
    k = self.f_hk_org(p, x)
    return np.where(x < self.hs, ks, modify * k)



# Fayer and Simmons model
import numpy as np


def init_model_fs(self):
    self.model['VGFS'] = self.model['Fayer-VG'] = self.model['vgfs'] = {
        'function': (self.vgfs, self.vgfs_k),
        'bound': self.bound_vgfs,
        'get_init': self.get_init_vgfs,
        'get_wrf': self.get_wrf_vgfs,
        'param': ['qs', 'qr', 'qa', 'a', 'm', 'he', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 9]
    }


def bound_vgfs(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_he, self.b_ks, self.b_p, self.b_q, self.b_r]


def vgfs(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vgfs_se(p, x) * (p[0] - p[1]) + p[1]


def vgfs_se(self, p, x):
    qs, qr, qa, a, m, he, q = p
    vg = self.vg_se([a, m, q], x)
    xi = np.where(x > 1, 1 - np.log(x) / np.log(he), 1)
    xisa = xi * qa / qs
    return xisa + (1 - xisa) * vg


def vgfs_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, qa, a, m, hm, ks, p, q, r = par
    if q != 1:  # Note that q should be 1
        return 1
    sa = qa / qs
    n = q / (1 - m)
    beta = 1
    gse = 1 - np.log(a / beta) / np.log(beta * hm) * \
        sa - sa  # 1 - gamma Se - Se

    def gamma(x, a, m, n, hm, sa, gse):
        w = 1 / (1 + (a * x)**n)
        wm = 1 / (1 + (a * hm)**n)

        def gs(w, m):
            s = 0
            for k in range(101):
                s += w**k / (m + 1 + k)
            return w**(m + 1) * s

        def h(w, m):
            return np.log(w) - m * w + m * (m - 1) * w * w / \
                4 - m * (m - 1) * (m - 2) * w**3 / 18

        def g(w, m):
            return np.where(w > 0.9, gs(0.9, m)
                            + h(0.1, m) - h(1 - w, m), gs(w, m))

        def f(w, m):
            return (1 - w)**m * (np.log((1 - w) / w) - 1 / m)

        def i3b(w, wm, m):
            return m * ((w - wm) + wm * (np.log(wm)) - w * np.log(w))

        w0 = max(10**(-10), wm)
        i1 = (1 - wm)**m - (1 - w)**m
        i2 = (1 / x - 1 / hm) / a + (1 - wm)**(m - 1) - (1 - w)**(m - 1)
        i3a = f(w0, m) - f(w, m) + g(1 - w, m) - g(1 - w0, m)
        i3 = np.where(w > w0, i3a + i3b(w0, wm, m), i3b(w, wm, m))
        result = i1 * a * gse + a * sa / np.log(hm) * (i2 + i3 / n)
        return result
    h0 = 0.025 / a
    hc = 10**(-7)  # Typical values for h_c range from 10^-7 to 10^-20

    def f(ah, n):
        return ah**(n - 1) * (np.log(ah) - 1 / (n - 1))

    gamma_0c = a * (gse + sa / n / np.log(hm)) * \
        ((a * h0)**(n - 1) - (a * hc)**(n - 1))
    gamma_0c += a * sa / np.log(hm) * (f(a * h0, n) - f(a * hc, n))
    gamma_max = gamma(h0, a, m, n, hm, sa, gse) + gamma_0c
    integral = gamma(x, a, m, n, hm, sa, gse) / gamma_max
    return ks * self.vgfs_se(par[:6] + [q], x)**p * integral**r


def get_init_vgfs(self, he):  # qa, a, m
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a, m, l2 = f.get_init_vgbcch()
    result = f.ini = (1 - w, a, m)
    f.set_model('vgfs', const=['qs=1', 'qr=0', [6, he], 'q=1'])
    f.optimize()
    if f.success:
        result = f.fitted
    return result


def get_wrf_vgfs(self, he):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    qa, a, m = f.get_init_vgfs(he)
    f.set_model('vgfs', const=['qr=0', [6, he], 'q=1'])
    qs = max(f.swrc[1])
    f.ini = (qs, qa, a, m)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], he, 1)
    return (qs, 0, *f.ini[1:], he, 1)



# tri-VG model
import numpy as np


def init_model_tri_vg(self):
    self.model['TV'] = self.model['tri-VG'] = self.model['vg3'] = {
        'function': (self.vg3, self.vg3_k),
        'bound': self.bound_vg3,
        'get_init': self.get_init_vg3,
        # ww2 = w2 / (1-w1)
        # w2 was transformed to make the bound independent of w1
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'ww2', 'a2', 'm2', 'a3', 'm3', 'Ks', 'p', 'q', 'r'],
        'k-only': [10, 11, 13],
        'sort_param': self.sort_param_vg3
    }


def bound_vg3(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_w1, self.b_a2, self.b_m, self.b_a2, self.b_m, self.b_ks, self.b_p, self.b_q, self.b_r]


def vg3(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vg3_se(p, x) * (p[0] - p[1]) + p[1]


def vg3_se(self, p, x):
    s1 = self.vg_se([p[3], p[4], p[10]], x)
    s2 = self.vg_se([p[6], p[7], p[10]], x)
    s3 = self.vg_se([p[8], p[9], p[10]], x)
    w1 = p[2]
    w2 = (1 - w1) * p[5]
    return w1 * s1 + w2 * s2 + (1 - w1 - w2) * s3


def vg3_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w1, a1, m1, ww2, a2, m2, a3, m3, ks, p, q, r = par
    w2 = (1 - w1) * ww2
    w1a1q = w1 * (a1 ** q)
    w2a2q = w2 * (a2 ** q)
    w3a3q = (1 - w1 - w2) * (a3 ** q)
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)
    s3 = self.vg_se([a3, m3, q], x)
    bunshi = w1a1q * (1 - (1 - s1**(1 / m1))**m1) + \
        w2a2q * (1 - (1 - s2**(1 / m2))**m2) + \
        w3a3q * (1 - (1 - s3**(1 / m3))**m3)
    bunbo = w1a1q + w2a2q + w3a3q
    return ks * self.vg3_se(par[:10] + [q], x)**p * (bunshi / bunbo)**r


def get_init_vg3(self, a1=0):  # w1, alpha1, m1, ww2, alpha2, m2, alpha3, m3
    # w2 = (1-w1) * ww2
    if a1 > 0:
        return self.get_init_vg3_fix_a1(a1)
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a1, m1, a3, m3 = f.get_init_vg2()
    if a1 < a3:
        a1, a3 = a3, a1
        m1, m3 = m3, m1
        w = 1 - w
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.b_a1 = self.b_a1
    f.b_a2 = self.b_a1
    f.b_a3 = self.b_a1
    if a1 < f.b_a1[0]:
        a1 = f.b_a1[0] * 1.0001
    if a1 > f.b_a1[1]:
        a1 = f.b_a1[1] * 0.9999
    f.set_model('vg3', const=[[1, 1], [2, 0], [9, a3], [10, m3], 'q=1'])
    r2 = 0
    for p in (0.05, 0.3, 0.6):
        f.ini = (w * p, a1, m1, w * (1 - p) / (1 - w * p), a1, m1)
        f.optimize()
        if r2 < f.r2_ht:
            fitted = f.fitted
            r2 = f.r2_ht
    f.set_model('vg3', const=[[1, 1], [2, 0], 'q=1'])
    f.ini = (*fitted, a3, m3)
    f.optimize()
    # sort
    w1, a1, m1, ww2, a2, m2, a3, m3 = f.fitted
    w2 = (1 - w1) * ww2
    param = ((w1, a1, m1), (w2, a2, m2), (1 - w1 - w2, a3, m3))
    ((w1, a1, m1), (w2, a2, m2), (w3, a3, m3)) = tuple(
        sorted(param, key=lambda x: x[1], reverse=True))
    f.fitted = w1, a1, m1, w2 / (1 - w1), a2, m2, a3, m3
    if f.success:
        ch = f.fitted
        ch_r2 = f.r2_ht
    else:
        ch = f.ini
        ch_r2 = f.f_r2_ht(f.ini, x, y)
    if len(x) < 10:
        return ch
    # Find a place to split to subfunctions
    try:
        c = np.polyfit(np.log10(x), y, 7)  # Fit with 7th degree
    except BaseException:
        return ch
    y_pred = np.polyval(c, np.log10(x))
    ss_res = np.sum((y - y_pred) ** 2)  # residual sum of squares
    ss_tot = np.sum((y - np.mean(y)) ** 2)  # total sum of squares
    r_squared = 1 - (ss_res / ss_tot)
    d = np.polyder(np.polyder(c))  # Second derivative
    split = []
    for root in np.roots(d):
        if root.imag == 0 and np.polyval(
                d, root + 0.01) < 0 and np.polyval(c, root) > 0.1:
            split.append(10**root.real)
    split.sort()
    if r_squared < 0.8 or len(split) < 2:
        return ch
    swrc = list(zip(*f.swrc))
    swrc_sort = sorted(swrc, key=lambda x: x[0])
    swrc = list(zip(*swrc_sort))
    h = np.array(swrc[0])
    t = np.array(swrc[1])
    i, j = np.searchsorted(h, split, side='right')[:2]
    if i < 4 or j - i < 3:
        return ch
    # Fit first subfunction
    f.swrc = (h[:i], t[:i] - t[i - 1])
    try:
        a1, m1 = f.get_init_vg()
    except BaseException:
        return ch
    if m1 > m_max:
        m1 = m_max
    # Fit second subfunction
    f.swrc = (h[i - 1:j], t[i - 1:j] - t[j - 1])
    try:
        a2, m2 = f.get_init_vg()
    except BaseException:
        return ch
    if m2 > m_max:
        m2 = m_max
    # Fit third subfunction
    f.swrc = (h[j - 1:], t[j - 1:])
    try:
        a3, m3 = f.get_init_vg()
    except BaseException:
        return ch
    if m3 > m_max:
        m3 = m_max
    # Final fitting
    w1 = 1 - t[i - 1]
    w2 = t[i - 1] - t[j - 1]
    ww2 = w2 / (1 - w1)
    f.set_model('vg3', const=['qs=1', 'qr=0', [3, w1], [6, ww2], 'q=1'])
    f.swrc = (h, t)
    f.ini = (a1, m1, a2, m2, a3, m3)
    f.optimize()
    a1, m1, a2, m2, a3, m3 = f.fitted
    f.set_model('vg3', const=['qs=1', 'qr=0', 'q=1'])
    f.ini = (w1, a1, m1, ww2, a2, m2, a3, m3)
    f.optimize()
    if not f.success:
        return ch
    if ch_r2 > f.r2_ht:
        return ch
    return f.fitted


def get_init_vg3_fix_a1(self, a1):  # w1, m1, ww2, alpha2, m2, alpha3, m3
    # w2 = (1-w1) * ww2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a2, m2, a3, m3 = f.get_init_vg2()
    if a2 < a3:
        a2, a3 = a3, a2
        m2, m3 = m3, m2
        w = 1 - w
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.set_model('vg3', const=[[1, 1], [2, 0],
                f'a1={a1}', f'a3={a3}', f'm3={m3}', 'q=1'])
    r2 = 0
    for p in (0.05, 0.3, 0.6):
        f.ini = (w * p, m2, w * (1 - p) / (1 - w * p), a2, m2)
        f.optimize()
        if r2 < f.r2_ht:
            fitted = f.fitted
            r2 = f.r2_ht
    f.set_model('vg3', const=[[1, 1], [2, 0], f'a1={a1}', 'q=1'])
    f.ini = (*fitted, a3, m3)
    f.optimize()
    if f.success:
        ch = f.fitted
    else:
        ch = f.ini
    return ch


def sort_param_vg3(self, param):
    """Sort fitted

    After f.set_model('tri-VG'), this function can be called with f.sort_param()

    Sort parameters of tri-VG model in the reverse order of alpha

    ww2 is converted to w2

    Input: w1, alpha1, m1, ww2, alpha2, m2, alpha3, m3

    Return: w1, alpha1, m1, w2, alpha2, m2, alpha3, m3
    """
    w1, a1, m1, ww2, a2, m2, a3, m3 = param
    w2 = (1 - w1) * ww2
    param = ((w1, a1, m1), (w2, a2, m2), (1 - w1 - w2, a3, m3))
    ((w1, a1, m1), (w2, a2, m2), (_, a3, m3)) = tuple(
        sorted(param, key=lambda x: x[1], reverse=True))
    return w1, a1, m1, w2, a2, m2, a3, m3



# BVV model
import numpy as np


def init_model_bvv(self):
    self.model['BVV'] = self.model['BC1VG2VG3'] = {
        'function': (self.bvv, self.bvv_k),
        'bound': self.bound_bvv,
        'get_init': self.get_init_bvv,
        # ww2 = w2 / (1-w1)
        # w2 was transformed to make the bound independent of w1
        'param': ['qs', 'qr', 'w1', 'hb1', 'l1', 'ww2', 'a2', 'm2', 'a3', 'm3', 'Ks', 'p', 'q', 'r'],
        'k-only': [10, 11, 13]
    }


def bound_bvv(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hb, self.b_lambda,
            self.b_w1, self.b_a2, self.b_m, self.b_a2, self.b_m, self.b_ks, self.b_p, self.b_q, self.b_r]


def bvv(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.bvv_se(p, x) * (p[0] - p[1]) + p[1]


def bvv_se(self, p, x):
    s1 = np.where(x < p[3], 1, (x / p[3]) ** (-p[4]))
    s2 = self.vg_se([p[6], p[7], p[10]], x)
    s3 = self.vg_se([p[8], p[9], p[10]], x)
    w1 = p[2]
    w2 = (1 - w1) * p[5]
    return w1 * s1 + w2 * s2 + (1 - w1 - w2) * s3


def bvv_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w1, hb1, l1, ww2, a2, m2, a3, m3, ks, p, q, r = par
    w2 = (1 - w1) * ww2
    w1a1q = w1 / hb1 ** q / (q / l1 + 1)
    w2a2q = w2 * (a2 ** q)
    w3a3q = (1 - w1 - w2) * (a3 ** q)
    s1 = np.where(x < hb1, 1, (x / hb1) ** (-l1 - q))
    s2 = self.vg_se([a2, m2, q], x)
    s3 = self.vg_se([a3, m3, q], x)
    bunshi = w1a1q * s1 + \
        w2a2q * (1 - (1 - s2**(1 / m2))**m2) + \
        w3a3q * (1 - (1 - s3**(1 / m3))**m3)
    bunbo = w1a1q + w2a2q + w3a3q
    return ks * self.bvv_se(par[:10] + [q], x)**p * (bunshi / bunbo)**r


def get_init_bvv(self, hb1=0):  # w1, hb1, l1, ww2, alpha2, m2, alpha3, m3
    # w2 = (1-w1) * ww2
    if hb1 > 0:
        return self.get_init_bvv_fix_hb1(hb1)
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, a1, m1, ww2, a2, m2, a3, m3 = f.get_init_vg3()
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.b_hb = self.b_hb
    if self.b_hb[0] > 0:
        f.b_a2 = f.b_a3 = (0, 1 / self.b_hb[0])
    f.set_model('BVV', const=[[1, 1], [2, 0], [3, w1], [6, ww2], [
                7, a2], [8, m2], [9, a3], [10, m3], 'q=1'])
    f.ini = (max(1 / a1, f.b_hb[0] * 1.001), 1 / (1 - m1) - 1)
    f.optimize()
    hb1, l1 = f.fitted
    f.set_model('BVV', const=[[1, 1], [2, 0], 'q=1'])
    f.ini = (w1, hb1, l1, ww2, a2, m2, a3, m3)
    f.optimize()
    return f.fitted


def get_init_bvv_fix_hb1(self, hb1):  # w1, l1, ww2, alpha2, m2, alpha3, m3
    # w2 = (1-w1) * ww2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, m1, ww2, a2, m2, a3, m3 = f.get_init_vg3(a1=1 / hb1)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.b_hb = self.b_hb
    if self.b_hb[0] > 0:
        f.b_a2 = f.b_a3 = (0, 1 / self.b_hb[0])
    f.set_model('BVV', const=[[1, 1], [2, 0], [3, w1], [4, hb1], [6, ww2], [
                7, a2], [8, m2], [9, a3], [10, m3], 'q=1'])
    f.ini = (1 / (1 - m1) - 1)
    f.optimize()
    l1 = f.fitted[0]
    f.set_model('BVV', const=[[1, 1], [2, 0], [4, hb1], 'q=1'])
    f.ini = (w1, l1, ww2, a2, m2, a3, m3)
    f.optimize()
    return f.fitted



# Pore size distribution


def f_pore(self, p, r, C=0.149, epsilon=1e-8):
    """
    Pore-size distribution function dθ/dr

    Note that figure should be drawn with normal scale.
    For using log-scale of pore-size, use f_pore_log.

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    r : numpy.ndarray
        pore size
    C : float
        constant for relationship h = C/r
        unit dependent. 0.149 cm^2
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/dr
    """
    h = C / r
    return -C / (r**2) * self.f_dtdh(p, h, epsilon=epsilon)


def f_pore_log(self, p, r, C=0.149, epsilon=1e-8):
    """
    Pore-size distribution function dθ/d(ln r)

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    r : numpy.ndarray
        pore size
    C : float
        constant for relationship h = C/r
        unit dependent. 0.149 cm^2
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/d(ln r)
    """
    h = C / r
    return -h * self.f_dtdh(p, h, epsilon=epsilon)


def f_dtdh(self, p, h, epsilon=1e-8):
    """
    dθ/dh calculated with numerical differentiation,
            using 4-point central difference formula

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    h : numpy.ndarray
        pressure head
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/dh
    """
    theta_values = [self.f_ht(p, h + k * epsilon) for k in [-2, -1, 1, 2]]
    return sum(
        k * theta for k, theta in zip([1, -8, 8, -1], theta_values)) / (12 * epsilon)


def find_pore_extreme(self, p, epsilon=1e-6):
    """
    Find inflection points of θ(h; p) on the log-h axis by solving d²θ/d(ln h)² = 0, then
    classify them into local maxima/minima and store the results in self.pore_maxima and
    self.pore_minima.

    Method:
    - Evaluate dθ/dln h = h * dθ/dh (d1_cdiff) using a high-order centered finite difference
    with a relative step (epsilon).
    - Evaluate d²θ/d(ln h)² (d2_cdiff) by differentiating d1_cdiff with the same scheme.
    - Build a geometric grid over [min(self.swrc[0]) / 2, max(self.swrc[0])] to detect sign
    changes (or near-zero crossings) of d²θ/d(ln h)² and form brackets.
    - Refine each bracket with Brent’s method (scipy.optimize.root_scalar) to locate roots.
    - Classify each root r as:
    - maximum if dθ/dln h(r) > dθ/dln h(1.1 r),
    - minimum if dθ/dln h(r) < dθ/dln h(1.1 r).

    Parameters:
    - p (array-like): Parameter vector consumed by self.f_ht.
    - epsilon (float, optional): Relative step for the centered finite-difference stencil
    (default: 1e-6). Smaller values increase round-off sensitivity; larger values increase
    discretization error.

    Requirements:
    - self.f_ht(p, h) must exist, accept scalar/array h, and be vectorized.
    - self.swrc must provide (h, θ) data with finite bounds to define the search interval.

    Returns:
    - None. Populates:
    - self.pore_maxima: list[float], locations h* of local maxima on the log-h axis.
    - self.pore_minima: list[float], locations h* of local minima on the log-h axis.

    Raises:
    - ValueError: If the h range inferred from self.swrc is invalid (non-finite or lo >= hi).

    Side effects:
    - Overwrites/creates self.pore_maxima and self.pore_minima.

    Notes:
    - Uses a 5-point-like centered finite difference for the first derivative on a relative
    step and applies it again to obtain the second derivative in log-scale.
    - Root finding uses a 512-point geometric grid, a scale-aware zero threshold, and
    fixed tolerances (xtol=rtol=1e-6, maxiter=100).
    - The number of minima/maxima depends on the model/data. Downstream plotting expects
    exactly two minima when pore segmentation is enabled.

    Example:
        self.find_pore_extreme(self.fitted)
    """
    import numpy as np
    from scipy.optimize import root_scalar

    def diff(f, p, h, epsilon=1e-8):
        h = np.asarray(h, dtype=float)
        t_m2 = f(p, h * (1 - 2 * epsilon))
        t_m1 = f(p, h * (1 - epsilon))
        t_p1 = f(p, h * (1 + epsilon))
        t_p2 = f(p, h * (1 + 2 * epsilon))
        return (t_m2 - 8 * t_m1 + 8 * t_p1 - t_p2) / (12 * epsilon * h)

    def d1_cdiff(p, h, epsilon=1e-6):
        """ Multiply h for converting to log scale """
        return h * diff(self.f_ht, p, h, epsilon)

    def d2_cdiff(p, h, epsilon=1e-6):
        return diff(d1_cdiff, p, h, epsilon)

    # Build initial grid only to form sign-change brackets
    grid_n = 512
    xtol = 1e-6
    rtol = 1e-6
    lo = min(self.swrc[0]) / 2
    hi = max(self.swrc[0])
    if not np.isfinite(lo) or not np.isfinite(hi) or lo >= hi:
        raise ValueError("Invalid h range.")
    xs = np.geomspace(lo, hi, grid_n)
    # Vectorized second derivative on the grid
    g2_vals = d2_cdiff(p, xs, epsilon)
    # Robust zero threshold (scale-aware)
    eps_zero = 1e-12 * (1.0 + np.nanmax(np.abs(g2_vals)))
    sign = np.sign(np.where(np.abs(g2_vals) < eps_zero, 0.0, g2_vals))
    # Collect brackets where sign changes or hits zero
    brackets = []
    for i in range(len(xs) - 1):
        s1, s2 = sign[i], sign[i + 1]
        if s1 == 0.0 or s1 * s2 < 0.0:
            brackets.append((xs[i], xs[i + 1]))
    # Scalar wrapper for root_scalar

    def g2_scalar(h):
        return float(d2_cdiff(p, np.array([h]), epsilon)[0])
    # Refine each bracket with Brent's method
    roots = []
    for a, b in brackets:
        try:
            sol = root_scalar(
                g2_scalar,
                bracket=[
                    a,
                    b],
                method='brentq',
                xtol=xtol,
                rtol=rtol,
                maxiter=100)
        except ValueError:
            continue
        if sol.converged:
            r = sol.root
            roots.append(r)
    # Retrieve maxima and minima of pore-size distribution
    self.pore_maxima, self.pore_minima = [], []
    for r in roots:
        inc = d1_cdiff(p, r, epsilon=1e-6) - \
            d1_cdiff(p, r * 1.1, epsilon=1e-6)
        if inc < 0:
            self.pore_maxima.append(r)
        elif inc > 0:
            self.pore_minima.append(r)



# VVP model
import numpy as np


def init_model_vvp(self):
    self.model['VVP'] = self.model['tri-PDI'] = {
        'function': (self.vvp, self.vvp_k),
        'bound': self.bound_vvp,
        'get_init': self.get_init_vvp,
        # ww2 = w2 / (1-w1)
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'ww2', 'a2', 'm2', 'he', 'Ks', 'p', 'q', 'r', 'a'],
        'k-only': [9, 10, 12, 13]
    }
    self.model['VVPS'] = {
        'function': (self.vvps, self.vvps_k),
        'bound': self.bound_vvps,
        'get_init': self.get_init_vvps,
        # ww2 = w2 / (1-w1)
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'ww2', 'a2', 'm2', 'hf', 'he', 'Ks', 'p', 'q', 'r', 'a'],
        'k-only': [10, 11, 13, 14]
    }

# VG1VG2PE3


def bound_vvp(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_w1, self.b_a2, self.b_m, self.b_he, self.b_ks, self.b_p, self.b_q, self.b_r, self.b_a]


def vvp(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vvp_se(p, x) * (p[0] - p[1]) + p[1]


def vvp_se(self, p, x):
    qs, qr, w1, a1, m1, ww2, a2, m2, he, q = p[:10]
    w2 = (1 - w1) * ww2
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)

    def L(h):
        return np.log(1 + a2 * h)
    s3 = np.where(x < 1 / a2, 1, (L(he) - L(x)) / (L(he) - np.log(2)))
    return w1 * s1 + w2 * s2 + (1 - w1 - w2) * s3


def vvp_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w1, a1, m1, ww2, a2, m2, he, ks, p, q, r, a = par
    # dual-VG component
    w2 = (1 - w1) * ww2
    w1a1q = w1 * (a1 ** q)
    w2a2q = w2 * (a2 ** q)
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)
    bunshi = w1a1q * (1 - (1 - s1**(1 / m1))**m1) + \
        w2a2q * (1 - (1 - s2**(1 / m2))**m2)
    bunbo = w1a1q + w2a2q
    k12 = self.vvp_se(par[:9] + [q], x)**p * (bunshi / bunbo)**r
    # Film component
    k3 = np.where(x < 1 / a2, 1, (a2 * x) ** (-a))
    bunshi = (1 - w1 - w2) * (a2 ** q)
    bunbo = w1a1q + (1 - w1) * (a2 ** q)
    omega = self.vvp_se(par[:9] + [q], 1 / a2)**p * (bunshi / bunbo)**r
    kr = k12 * (1 - omega) + k3 * omega
    return ks * kr


def get_init_vvp(self, he):  # w1, a1, m1, ww2, a2, m2
    # w2 = (1-w1) * ww2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, a1, m1, ww2, a2, m2, a3, m3 = f.get_init_vg3()
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    w2, hm, sigma2 = f.get_init_pk(he)
    ww2 = w2 / (1 - w1)
    if ww2 > 1:
        ww2 = 0.99
    a2 = 1 / hm
    n2 = sigma2 ** (-1.25) / 1.2 + 1
    m2 = 1 - 1 / n2
    f.set_model('VVP', const=[[1, 1], [2, 0], [9, he], 'q=1'])
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.b_a1 = self.b_a1
    f.b_a2 = self.b_a1
    if a1 < f.b_a1[0]:
        a1 = f.b_a1[0] * 1.0001
    if a1 > f.b_a1[1]:
        a1 = f.b_a1[1] * 0.9999
    if a2 < f.b_a2[0]:
        a2 = f.b_a2[0] * 1.0001
    if a2 > f.b_a2[1]:
        a2 = f.b_a2[1] * 0.9999
    if m1 > m_max:
        m1 = m_max * 0.9999
    if m2 > m_max:
        m2 = m_max * 0.9999
    f.ini = (w1, a1, m1, ww2, a2, m2)
    f.optimize()
    return f.fitted

# VG1VG2PE3, separate hf


def bound_vvps(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_w1, self.b_a2, self.b_m, self.b_hb, self.b_he, self.b_ks, self.b_p, self.b_q, self.b_r, self.b_a]


def vvps(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vvps_se(p, x) * (p[0] - p[1]) + p[1]


def vvps_se(self, p, x):
    qs, qr, w1, a1, m1, ww2, a2, m2, hf, he, q = p[:11]
    w2 = (1 - w1) * ww2
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)

    def L(h):
        return np.log(1 + h / hf)
    s3 = np.where(x < hf, 1, (L(he) - L(x)) / (L(he) - np.log(2)))
    return w1 * s1 + w2 * s2 + (1 - w1 - w2) * s3


def vvps_k(self, p, x):
    # Requires update as vvp_k
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w1, a1, m1, ww2, a2, m2, hf, he, ks, p, q, r, a = par
    w2 = (1 - w1) * ww2
    w1a1q = w1 * (a1 ** q)
    w2a2q = w2 * (a2 ** q)
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)
    bunshi = w1a1q * (1 - (1 - s1**(1 / m1))**m1) + \
        w2a2q * (1 - (1 - s2**(1 / m2))**m2)
    bunbo = w1a1q + w2a2q
    k12 = self.vvps_se(par[:10] + [q], x)**p * (bunshi / bunbo)**r
    k3 = np.where(x < hf, 1, (x / hf) ** (-a))
    w3b3 = (1 - w1 - w2) / (hf ** q)
    bunbo = w1a1q + w2a2q + w3b3
    omega = self.vvps_se(par[:10] + [q], hf)**p * (w3b3 / bunbo)**r
    return ks * (k12 * (1 - omega) + k3 * omega)


def get_init_vvps(self, he):  # w1, a1, m1, ww2, a2, m2, hf
    # w2 = (1-w1) * ww2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, a1, m1, ww2, a2, m2, a3, m3 = f.get_init_vg3()
    f.set_model('VVPS', const=[[1, 1], [2, 0], [9, he], 'q=1'])
    f.ini = (w1, a1, m1, ww2, a2, m2, 1 / a3)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.b_a1 = self.b_a1
    f.b_a2 = self.b_a1
    f.optimize()
    return f.fitted



# KO1BC2 model
import numpy as np


def init_model_kobc(self):
    self.model['KOBC'] = self.model['KO1BC2'] = self.model['KB'] = self.model['kobc'] = {
        'function': (self.kobc, self.kobc_k),
        'bound': self.bound_kobc,
        'get_init': self.get_init_kobc,
        'get_wrf': self.get_wrf_kobc,
        'param': ['qs', 'qr', 'w1', 'hm1', 'sigma1', 'hb2', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [7, 8, 9, 10]
    }
    self.model['KOBCIP'] = self.model['KO1BC2-IP'] = self.model['KB-IP'] = self.model['kobcp2'] = {
        'function': (self.kobc, self.kobcp2_k),
        'bound': self.bound_kobcp2,
        'get_init': self.get_init_kobc,
        'get_wrf': self.get_wrf_kobc,
        'param': ['qs', 'qr', 'w1', 'hm1', 'sigma1', 'hb2', 'l2', 'Ks', 'p1', 'p2', 'q'],
        'k-only': [7, 8, 9, 10]
    }
    self.model['KOBCCH'] = self.model['KO1BC2-CH'] = self.model['KBC'] = self.model['kobcch'] = {
        'function': (self.kobcch, self.kobcch_k),
        'bound': self.bound_kobcch,
        'get_init': self.get_init_kobcch,
        'get_wrf': self.get_wrf_kobcch,
        'param': ['qs', 'qr', 'w1', 'hm', 'sigma1', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 8, 9]
    }
    self.model['KOBCCHIP'] = self.model['KO1BC2-CH-IP'] = self.model['KBC-IP'] = self.model['kobcchp2'] = {
        'function': (self.kobcch, self.kobcchp2_k),
        'bound': self.bound_kobcchp2,
        'get_init': self.get_init_kobcch,
        'get_wrf': self.get_wrf_kobcch,
        'param': ['qs', 'qr', 'w1', 'hm', 'sigma1', 'l2', 'Ks', 'p1', 'p2', 'q'],
        'k-only': [6, 7, 8, 9]
    }
    self.model['kobcchca'] = {
        'function': (self.kobcch, self.kobcchca_k),
        'bound': self.bound_kobcch,
        'get_init': self.get_init_kobcch,
        'get_wrf': self.get_wrf_kobcch,
        'param': ['qs', 'qr', 'w1', 'hm', 'sigma1', 'l2', 'Ks', 'p', 'a', 'r'],
        'k-only': [6, 7, 8, 9]
    }


def bound_kobc(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_hb2, self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def kobc(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.kobc_se(p, x) * (p[0] - p[1]) + p[1]


def kobc_se(self, p, x):
    qs, qr, w, hm1, sigma1, hb2, l2 = p
    s1 = self.ln_se([hm1, sigma1], x)
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2))
    return w * s1 + (1 - w) * s2


def kobc_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, hm1, sigma1, hb2, l2, ks, p, q, r = par
    w1b1 = w * (hm1 ** (-q)) * np.exp((q * sigma1)**2 / 2)
    w2b2 = (1 - w) / hb2 ** q / (q / l2 + 1)
    s1 = 1 - norm.cdf(np.log(x / hm1) / sigma1 + q * sigma1)
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.kobc_se(par[:7], x)**p * (bunshi / bunbo)**r


def get_init_kobc(self):  # w1, hm1, sigma1, hb2, l2
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, hm, sigma1, l2 = f.get_init_kobcch()
    f.set_model('kobc', const=[[1, 1], [2, 0]])
    f.ini = (w1, hm, sigma1, hm, l2)
    f.optimize()
    if f.success:
        ch = f.fitted
        ch_r2 = f.r2_ht
    else:
        ch = f.ini
        ch_r2 = f.f_r2_ht(f.ini, x, y)
    if len(x) < 6:
        return ch
    swrc = list(zip(*f.swrc))
    swrc_sort = sorted(swrc, key=lambda x: x[0])
    swrc = list(zip(*swrc_sort))
    h = np.array(swrc[0])
    t = np.array(swrc[1])
    t_med = (max(t) + min(t)) / 2
    for i in range(len(h)):
        if t[i] < t_med:
            break
    if i < 3:
        i = 3
    if i > len(h) - 3:
        i = len(h) - 3
    f.swrc = (h[:i], t[:i] - t[i])
    try:
        hm1, sigma1 = f.get_init_ko()
    except BaseException:
        return ch
    f.swrc = (h[i:], t[i:])
    try:
        hb2, l2 = f.get_init_bc()
    except BaseException:
        return ch
    f.set_model('kobc', const=['qs=1', 'qr=0'])
    f.swrc = (h, t)
    f.ini = (1 - t[i], hm1, sigma1, hb2, l2)
    f.optimize()
    if not f.success:
        return ch
    if ch_r2 > f.r2_ht:
        return ch
    return f.fitted


def get_wrf_kobc(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w1, hm1, sigma1, hb2, l2 = f.get_init_kobc()
    f.set_model('kobc', const=['qr=0'])
    f.ini = (max(f.swrc[1]), w1, hm1, sigma1, hb2, l2)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:])
    qs, qr, w1, h1, s1, l2 = f.get_wrf_kobcch()
    return (qs, qr, w1, h1, s1, h1, l2)

# KO1BC2 model with r=1 and independent p1, p2


def bound_kobcp2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_hb2, self.b_lambda2, self.b_ks, self.b_p, self.b_p, self.b_q]


def kobcp2_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, hm1, sigma1, hb2, l2, ks, p1, p2, q = par
    w1b1 = w * (hm1 ** (-q)) * np.exp((q * sigma1)**2 / 2)
    w2b2 = (1 - w) / (hb2 ** q) / (q / l2 + 1)
    w1a1 = w1b1 * (1 - norm.cdf(np.log(x / hm1) / sigma1 + q * sigma1))
    w2a2 = w2b2 * (np.where(x < hb2, 1, (x / hb2) ** (-l2 - q)))
    # Seki et al. (2021) eq.(19)
    # s1 = 1 - norm.cdf(np.log(x / hm1)/sigma1)
    # s2 = np.where(x < hb2, 1, (x/hb2) ** (-l2))
    # bunshi = s1**p1 * w1a1 + s2**p2 * w2a2
    # Corrected from Si(h) to Se(h)
    se = self.kobc_se(par[:7], x)
    bunshi = se**p1 * w1a1 + se**p2 * w2a2  # Corrected
    bunbo = w1b1 + w2b2
    return ks * bunshi / bunbo

# KO1BC2-CH model


def bound_kobcch(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def kobcch(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.kobcch_se(p, x) * (p[0] - p[1]) + p[1]


def kobcch_se(self, p, x):
    qs, qr, w, h, sigma, l2 = p
    s1 = self.ln_se([h, sigma], x)
    s2 = np.where(x < h, 1, (x / h) ** (-l2))
    return w * s1 + (1 - w) * s2


def kobcch_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, h, sigma, l2, ks, p, q, r = par
    w1b1 = w * (h ** (-q)) * np.exp((q * sigma)**2 / 2)
    w2b2 = (1 - w) / h ** q / (q / l2 + 1)
    s1 = 1 - norm.cdf(np.log(x / h) / sigma + q * sigma)
    s2 = np.where(x < h, 1, (x / h) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.kobcch_se(par[:6], x)**p * (bunshi / bunbo)**r


def get_init_kobcch(self):  # w1, hm, sigma1, l2
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, a, m1, l2 = f.get_init_vgbcch()
    if w1 < 0.1 or w1 > 0.9:
        w1, a, m1, m2 = f.get_init_vg2ch()
        n2 = 1 / (1 - m2)
        l2 = n2 - 1
    f.set_model('kobcch', const=[[1, 1], [2, 0]])
    if l2 < 0.01:
        l2 = 0.01
    if l2 > 3:
        l2 = 2.99
    n1 = 1 / (1 - m1)
    sigma1 = 1.2 * (n1 - 1)**(-0.8)
    if sigma1 > 2.99:
        sigma1 = 2.99
    ini = f.ini = (w1, 1 / a, sigma1, l2)
    f.b_sigma = (0, 3)
    f.b_lambda2 = (0.001, 3)
    f.optimize()
    if f.success:
        return f.fitted
    h, sigma = f.get_init_ln()
    if sigma > 2.99:
        sigma = 2.99
    f.ini = ([0.2, 0.8], [h, 1 / a], [sigma, ], [l2, ])
    f.optimize()
    if f.success:
        return f.fitted
    return ini


def get_wrf_kobcch(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w1, h1, s1, l2 = f.get_init_kobcch()
    f.set_model('kobcch', const=['qr=0'])
    qs = max(f.swrc[1])
    f.ini = (qs, w1, h1, s1, l2)
    f.b_sigma = (0, 4)
    f.optimize()
    if f.success:
        import copy
        f2 = copy.deepcopy(f)
        f.b_sigma = (0, np.inf)
        f.optimize()
        if f.success and f.r2_ht > f2.r2_ht:
            return (f.fitted[0], 0, *f.fitted[1:])
        else:
            return (f2.fitted[0], 0, *f2.fitted[1:])
    f.b_qs = (qs * 0.95, qs * 1.5)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:])
    hb, hc, l1, l2 = f.get_init_bc2()
    w = 1 / (1 + (hc / hb)**(l2 - l1))
    s1 = 1.2 * l1**(-0.8)
    if s1 > 4:
        s1 = 4
    f.ini = (qs, w, hb, s1, l2)
    f.optimize()
    return (qs, 0, *f.ini[1:])

# KO1BC2-CH model with r=1 and independent p1, p2


def bound_kobcchp2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_lambda2, self.b_ks, self.b_p, self.b_p, self.b_q]


def kobcchp2_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, h, sigma, l2, ks, p1, p2, q = par
    w1b1 = w * (h ** (-q)) * np.exp((q * sigma)**2 / 2)
    w2b2 = (1 - w) / h ** q / (q / l2 + 1)
    w1a1 = w1b1 * (1 - norm.cdf(np.log(x / h) / sigma + q * sigma))
    w2a2 = w2b2 * (np.where(x < h, 1, (x / h) ** (-l2 - q)))
    # s1 = 1 - norm.cdf(np.log(x / h)/sigma)
    # s2 = np.where(x < h, 1, (x/h) ** (-l2))
    # bunshi = s1**p1 * w1a1 + s2**p2 * w2a2 # Seki et al. (2021) eq.(19)
    # Corrected from Si(h) to Se(h)
    se = self.kobcch_se(par[:6], x)
    bunshi = se**p1 * w1a1 + se**p2 * w2a2
    bunbo = w1b1 + w2b2
    return ks * bunshi / bunbo

# KO1BC2-CH model, constant a


def kobcchca_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, h, sigma, l2, ks, p, a, r = par
    q = (a - p * l2) / r - l2
    w1b1 = w * (h ** (-q)) * np.exp((q * sigma)**2 / 2)
    w2b2 = (1 - w) / h ** q / (q / l2 + 1)
    s1 = 1 - norm.cdf(np.log(x / h) / sigma + q * sigma)
    s2 = np.where(x < h, 1, (x / h) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.kobcch_se(par[:6], x)**p * (bunshi / bunbo)**r
